/* Jonathan E. Shopiro
   An Example of Multiple Inheritance in C++:
   A Model of the Iostream Library
*/


extern int open (const char*, int, ...);
extern int close (int);
extern int creat (const char*, int);
extern long lseek (int, long, int);
extern int read (int, void*, unsigned int);
extern int write (int, const void*, unsigned int);
#define	O_RDONLY	0
#define O_WRONLY	1
#define O_RDWR		2
#define	O_CREAT	0x0100	

const int bufsiz = 17; // for debugging

class istream;
class ostream;

class streambuf {
	friend istream;
	friend ostream;
protected:
	// encapsulates buffer management
	char buf[bufsiz]; // the character buffer
	char* x_gptr; // get pointer
	char* x_pptr; // put pointer
	bool is_writing; // buffer must be written
	virtual void overflow(int c) = 0; // send full buffer
	virtual void underflow() = 0; // refill empty buffer
	int sbumpc() { // get a character
		if (x_gptr >= &buf[bufsiz]) underflow() ;
		return *x_gptr++; 
	}

	void sputc(int c) { // put a character
		if (x_pptr >= &buf[bufsiz]) overflow(c) ;
		else *x_pptr++ = c; 
	}
public:
	streambuf() : is_writing(false), x_gptr(&buf[bufsiz]), 
		x_pptr(&buf[bufsiz]) 
	{}
	virtual ~streambuf() {}
};

class ios { // a connection to a streambuf
public :
	enum open_mode { in=1, out=2 } ;
	ios (streambuf*p) : bp(p) {}
	virtual ~ios() { delete bp; }
protected:
	streambuf * bp;
	ios() {} // never called
};

class istream : public virtual ios { // input stream class
public:
	istream& operator>>(char& c) { 
		c = bp->sbumpc();
		return *this; 
	}
};

class ostream : public virtual ios { // output stream class
public:
	ostream& operator<<(char c) { 
		bp->sputc(c);
		return *this; 
	}
};

class iostream : public istream, public ostream {};

class fstreambase;
class fstream;
class ifstream;
class ofstream;

class filebuf : public streambuf {
	// a streambuf for files
	friend fstreambase;
	friend fstream;
	friend ifstream;
	friend ofstream;
	int fd;
	void overflow(int) ;
	void underflow() ;
	void seek(long offset, int whence);
	void flush() { if (is_writing) write(fd, buf, x_pptr-buf); }
	filebuf(const char *name, int om);
	~filebuf() ;
};

class fstreambase : virtual public ios {
protected:
	fstreambase() { }
public:
	// the member streambuf: :bp is here known to point to a filebuf
	void seek (long offset, int whence) {
		( (filebuf*)bp)->seek (offset, whence) ; 
	}
};


class ifstream : public fstreambase, public istream {
public:
	ifstream(const char* name) : ios(new filebuf(name, ios::in)) {}
};

class ofstream : public fstreambase, public ostream {
public:
	ofstream(const char* name) : ios(new filebuf(name, ios::out)) {}
};

class fstream : public fstreambase, public iostream {
public:
	fstream(const char* name) : ios(new filebuf(name,ios::in | ios::out)) {}
};
